@@ -34,14 +34,16 @@
 #include <ctype.h>
 #include <unistd.h>
 #include <sys/mman.h>
+#include <sys/types.h>
 
 #include <wayland-client.h>
-#if HAVE_HUMAN_FORMAT_MODIFIER
-#include "xf86drm.h"
-#include "drm_fourcc.h"
+#if HAVE_LIBDRM
+#include <xf86drm.h>
+#include <drm_fourcc.h>
 #endif
 
 #include "presentation-time-client-protocol.h"
+#include "drm-lease-v1-client-protocol.h"
 #include "linux-dmabuf-unstable-v1-client-protocol.h"
 #include "tablet-unstable-v2-client-protocol.h"
 #include "xdg-output-unstable-v1-client-protocol.h"
@@ -259,9 +261,30 @@ struct presentation_info {
 	clockid_t clk_id;
 };
 
+struct drm_lease_connector {
+	struct wl_list link;
+
+	struct wp_drm_lease_connector_v1 *connector;
+
+	uint32_t id;
+	char *name, *description;
+};
+
+struct drm_lease_device_info {
+	struct global_info global;
+	struct wl_list link;
+
+	struct wp_drm_lease_device_v1 *device;
+
+	char *path;
+
+	struct wl_list connectors;
+};
+
 struct wayland_info {
 	struct wl_display *display;
 	struct wl_registry *registry;
+	char *interface;
 
 	struct wl_list infos;
 	bool roundtrip_needed;
@@ -478,6 +501,16 @@ print_shm_info(void *data)
 	printf("\n");
 }
 
+static void
+print_format(uint32_t format)
+{
+	char str[5];
+
+	fourcc2str(format, str, sizeof(str));
+	printf("0x%08x = '%s'",
+	       format, str);
+}
+
 static void
 print_format_and_modifier(uint32_t format, uint64_t modifier)
 {
@@ -485,7 +518,7 @@ print_format_and_modifier(uint32_t format, uint64_t modifier)
 
 	fourcc2str(format, str, sizeof(str));
 
-#if HAVE_HUMAN_FORMAT_MODIFIER
+#if HAVE_LIBDRM
 	char *name = drmGetFormatModifierName(modifier);
 	char *vendor = drmGetFormatModifierVendor(modifier);
 	if (!vendor)
@@ -509,6 +542,34 @@ print_format_and_modifier(uint32_t format, uint64_t modifier)
 #endif
 }
 
+static void
+print_devid(dev_t devid)
+{
+	printf("0x%" PRIX64, (uint64_t) devid);
+
+#if HAVE_LIBDRM
+	drmDevice *dev = NULL;
+	if (drmGetDeviceFromDevId(devid, 0, &dev) != 0) {
+		perror("drmGetDeviceFromDevId failed");
+		return;
+	}
+
+	if (!(dev->available_nodes & (1 << DRM_NODE_PRIMARY))) {
+		/* Should never happen */
+		fprintf(stderr, "Device has no primary node\n");
+		return;
+	}
+
+	printf(" (%s", dev->nodes[DRM_NODE_PRIMARY]);
+	if (dev->available_nodes & (1 << DRM_NODE_RENDER)) {
+		printf(" or %s", dev->nodes[DRM_NODE_RENDER]);
+	}
+	printf(")");
+
+	drmFreeDevice(&dev);
+#endif
+}
+
 static void
 print_linux_dmabuf_info(void *data)
 {
@@ -519,7 +580,8 @@ print_linux_dmabuf_info(void *data)
 	print_global_info(data);
 
 	if (dmabuf->main_device) {
-		printf("\tmain device: 0x%" PRIX64, (uint64_t) dmabuf->main_device);
+		printf("\tmain device: ");
+		print_devid(dmabuf->main_device);
 
 		if (dmabuf->format_map == MAP_FAILED) {
 			printf("\tError: failed to map format table\n");
@@ -535,8 +597,8 @@ print_linux_dmabuf_info(void *data)
 		wl_list_for_each(tranche, &dmabuf->tranches, link) {
 			printf("\n\ttranche");
 
-			printf("\n\t\ttarget device: 0x%" PRIX64,
-			       (uint64_t) tranche->target_device);
+			printf("\n\t\ttarget device: ");
+			print_devid(tranche->target_device);
 			printf("\n\t\tflags: ");
 			if (tranche->flags == 0)
 				printf("none");
@@ -544,7 +606,7 @@ print_linux_dmabuf_info(void *data)
 				printf("scanout");
 			printf("\n");
 
-#if HAVE_HUMAN_FORMAT_MODIFIER
+#if HAVE_LIBDRM
 			printf("\t\tformats (fourcc) and modifiers (names):");
 #else
 			printf("\t\tformats (fourcc) and modifiers:");
@@ -559,8 +621,8 @@ print_linux_dmabuf_info(void *data)
 		}
 
 		munmap(dmabuf->format_map, dmabuf->format_size);
-	} else {
-#if HAVE_HUMAN_FORMAT_MODIFIER
+	} else if (dmabuf->global.version > 2) {
+#if HAVE_LIBDRM
 		printf("\tformats (fourcc) and modifiers (names):");
 #else
 		printf("\tformats (fourcc) and modifiers:");
@@ -570,6 +632,12 @@ print_linux_dmabuf_info(void *data)
 			printf("\n\t");
 			print_format_and_modifier(modifier->format, modifier->modifier);
 		}
+	} else {
+		printf("\tformats (fourcc):");
+		wl_list_for_each(modifier, &dmabuf->modifiers, link) {
+			printf("\n\t");
+			print_format(modifier->format);
+		}
 	}
 
 	printf("\n");
@@ -1682,7 +1750,15 @@ add_shm_info(struct wayland_info *info, uint32_t id, uint32_t version)
 static void
 linux_dmabuf_handle_format(void *data, struct zwp_linux_dmabuf_v1 *zwp_linux_dmabuf_v1, uint32_t format)
 {
-	/* This is a deprecated event, don’t use it. */
+	struct linux_dmabuf_info *dmabuf = data;
+	struct linux_dmabuf_modifier *linux_dmabuf_modifier;
+
+	if (dmabuf->global.version > 2)
+		return;
+
+	linux_dmabuf_modifier = xzalloc(sizeof *linux_dmabuf_modifier);
+	wl_list_insert(&dmabuf->modifiers, &linux_dmabuf_modifier->link);
+	linux_dmabuf_modifier->format = format;
 }
 
 static void
@@ -1719,6 +1795,9 @@ destroy_linux_dmabuf_info(void *data)
 		free(tranche);
 	}
 
+	if (dmabuf->feedback)
+		zwp_linux_dmabuf_feedback_v1_destroy(dmabuf->feedback);
+
 	zwp_linux_dmabuf_v1_destroy(dmabuf->dmabuf);
 }
 
@@ -1807,13 +1886,14 @@ add_linux_dmabuf_info(struct wayland_info *info, uint32_t id, uint32_t version)
 	wl_list_init(&dmabuf->tranches);
 	wl_list_init(&dmabuf->modifiers);
 
-	if (version == 3) {
+	if (version <= 3) {
 		dmabuf->dmabuf = wl_registry_bind(info->registry,
-		                                  id, &zwp_linux_dmabuf_v1_interface, 3);
+		                                  id, &zwp_linux_dmabuf_v1_interface,
+		                                  MIN(version, 3));
 		zwp_linux_dmabuf_v1_add_listener(dmabuf->dmabuf, &linux_dmabuf_listener, dmabuf);
 
 		info->roundtrip_needed = true;
-	} else if (version >= 4) {
+	} else {
 		dmabuf->dmabuf = wl_registry_bind(info->registry,
 		                                  id, &zwp_linux_dmabuf_v1_interface, 4);
 		dmabuf->feedback = zwp_linux_dmabuf_v1_get_default_feedback(dmabuf->dmabuf);
@@ -2030,7 +2110,160 @@ add_presentation_info(struct wayland_info *info, uint32_t id, uint32_t version)
 }
 
 static void
-destroy_global_info(void *data)
+drm_lease_connector_name(void *data,
+		struct wp_drm_lease_connector_v1 *connector, const char *name)
+{
+	struct drm_lease_connector *con = data;
+	con->name = strdup(name);
+}
+
+static void
+drm_lease_connector_description(void *data,
+		struct wp_drm_lease_connector_v1 *connector, const char *desc)
+{
+	struct drm_lease_connector *con = data;
+	con->description = strdup(desc);
+}
+
+static void
+drm_lease_connector_id(void *data, struct wp_drm_lease_connector_v1 *connector,
+		uint32_t id)
+{
+	struct drm_lease_connector *con = data;
+	con->id = id;
+}
+
+static void
+drm_lease_connector_done(void *data,
+		struct wp_drm_lease_connector_v1 *connector)
+{
+	/* no-op */
+}
+
+static void
+drm_lease_connector_withdrawn(void *data,
+		struct wp_drm_lease_connector_v1 *connector)
+{
+	/* no-op */
+}
+
+static const struct wp_drm_lease_connector_v1_listener drm_lease_connector_listener = {
+	.name = drm_lease_connector_name,
+	.description = drm_lease_connector_description,
+	.connector_id = drm_lease_connector_id,
+	.done = drm_lease_connector_done,
+	.withdrawn = drm_lease_connector_withdrawn,
+};
+
+static void
+drm_lease_device_drm_fd(void *data, struct wp_drm_lease_device_v1 *device,
+		int32_t fd)
+{
+	struct drm_lease_device_info *dev = data;
+#if HAVE_LIBDRM
+	dev->path = drmGetDeviceNameFromFd(fd);
+#else
+	dev->path = NULL;
+#endif
+	close(fd);
+}
+
+static void
+drm_lease_device_connector(void *data, struct wp_drm_lease_device_v1 *device,
+		struct wp_drm_lease_connector_v1 *connector)
+{
+	struct drm_lease_device_info *dev = data;
+
+	struct drm_lease_connector *conn = xzalloc(sizeof *conn);
+
+	conn->connector = connector;
+
+	wp_drm_lease_connector_v1_add_listener(connector,
+		&drm_lease_connector_listener, conn);
+
+	wl_list_insert(&dev->connectors, &conn->link);
+}
+
+static void
+drm_lease_device_done(void *data, struct wp_drm_lease_device_v1 *device)
+{
+	/* no-op */
+}
+
+static void
+drm_lease_device_released(void *data, struct wp_drm_lease_device_v1 *device)
+{
+	/* no-op */
+}
+
+static const struct wp_drm_lease_device_v1_listener drm_lease_device_listener = {
+	.drm_fd = drm_lease_device_drm_fd,
+	.connector = drm_lease_device_connector,
+	.done = drm_lease_device_done,
+	.released = drm_lease_device_released,
+};
+
+static void
+destroy_drm_lease_device_info(void *info)
+{
+	struct drm_lease_device_info *dev = info;
+
+	free(dev->path);
+
+	wp_drm_lease_device_v1_destroy(dev->device);
+
+	struct drm_lease_connector *conn, *tmp;
+	wl_list_for_each_safe(conn, tmp, &dev->connectors, link) {
+		wl_list_remove(&conn->link);
+
+		wp_drm_lease_connector_v1_destroy(conn->connector);
+
+		free(conn->name);
+		free(conn->description);
+		free(conn);
+	}
+}
+
+static void
+print_drm_lease_device_info(void *info)
+{
+	print_global_info(info);
+	struct drm_lease_device_info *dev = info;
+
+	printf("\tpath: %s\n", dev->path);
+
+	struct drm_lease_connector *conn;
+	wl_list_for_each(conn, &dev->connectors, link) {
+		printf("\tconnector:\n");
+		printf("\t\tid: %u\n", conn->id);
+		printf("\t\tname: %s\n", conn->name);
+		printf("\t\tdescription: %s\n", conn->description);
+	}
+}
+
+static void
+add_drm_lease_device_v1_info(struct wayland_info *info, uint32_t id,
+		uint32_t version)
+{
+	struct drm_lease_device_info *dev = xzalloc(sizeof *dev);
+
+	init_global_info(info, &dev->global, id,
+		wp_drm_lease_device_v1_interface.name, version);
+	dev->global.print = print_drm_lease_device_info;
+	dev->global.destroy = destroy_drm_lease_device_info;
+
+	dev->device = wl_registry_bind(info->registry, id,
+		&wp_drm_lease_device_v1_interface, 1);
+	wp_drm_lease_device_v1_add_listener(dev->device,
+		&drm_lease_device_listener, dev);
+
+	wl_list_init(&dev->connectors);
+
+	info->roundtrip_needed = true;
+}
+
+static void
+destroy_global_info(void *info)
 {
 }
 
@@ -2065,6 +2298,8 @@ global_handler(void *data, struct wl_registry *registry, uint32_t id,
 		add_tablet_v2_info(info, id, version);
 	else if (!strcmp(interface, zxdg_output_manager_v1_interface.name))
 		add_xdg_output_manager_v1_info(info, id, version);
+	else if (!strcmp(interface, wp_drm_lease_device_v1_interface.name))
+		add_drm_lease_device_v1_info(info, id, version);
 	else
 		add_global_info(info, id, interface, version);
 }
@@ -2079,13 +2314,22 @@ static const struct wl_registry_listener registry_listener = {
 	global_remove_handler
 };
 
-static void
-print_infos(struct wl_list *infos)
+static bool
+should_print_info(struct wayland_info *wayland_info, struct global_info *global)
 {
+	return !wayland_info->interface ||
+		strstr(global->interface, wayland_info->interface);
+}
+
+static void
+print_infos(struct wayland_info *wayland_info)
+{
+	struct wl_list *infos = &wayland_info->infos;
 	struct global_info *info;
 
 	wl_list_for_each(info, infos, link)
-		info->print(info);
+		if (should_print_info(wayland_info, info))
+			info->print(info);
 }
 
 static void
@@ -2107,11 +2351,48 @@ destroy_infos(struct wl_list *infos)
 		destroy_info(info);
 }
 
+static void
+print_usage(const char *name, int exit_status)
+{
+	fprintf(stderr, "Usage: %s [options...] \n", name);
+	fprintf(stderr, "\n");
+	fprintf(stderr, " -h|--help: Print this help and exit\n");
+	fprintf(stderr, " -i|--interface <interface>: Only print information about the"
+			" Wayland globals containing <interface> in their name\n");
+	fprintf(stderr, "\n");
+	exit(exit_status);
+}
+
+static void
+parse_cmdline(struct wayland_info *info, int argc, char **argv)
+{
+	int i;
+
+	info->interface = NULL;
+
+	for (i = 1; i < argc; i++) {
+		if (!strcmp(argv[i], "--help") || !strcmp(argv[i], "-h")) {
+			print_usage(argv[0], EXIT_SUCCESS);
+		} else if (!strcmp(argv[i], "--interface") || !strcmp(argv[i], "-i")) {
+			if (i + 1 >= argc) {
+				fprintf(stderr, "Missing interface name\n");
+				print_usage(argv[0], EXIT_FAILURE);
+			}
+			info->interface = strdup(argv[++i]);
+		} else {
+			fprintf(stderr, "Unknown option %s\n", argv[i]);
+			print_usage(argv[0], EXIT_FAILURE);
+		}
+	}
+}
+
 int
 main(int argc, char **argv)
 {
 	struct wayland_info info;
 
+	parse_cmdline(&info, argc, argv);
+
 	info.display = wl_display_connect(NULL);
 	if (!info.display) {
 		fprintf(stderr, "failed to create display: %s\n",
@@ -2133,7 +2414,7 @@ main(int argc, char **argv)
 		wl_display_roundtrip(info.display);
 	} while (info.roundtrip_needed);
 
-	print_infos(&info.infos);
+	print_infos(&info);
 	destroy_infos(&info.infos);
 
 	wl_registry_destroy(info.registry);
